from typing import Any, Dict, Iterable, List, Optional, Union

from ...pipeline.normalizers import LowercaseStrings
from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...schema.indexes import FieldIndex, KeyIndex
from ...schema.schema import (
    GraphObjectShape,
    GraphObjectType,
    KnownTypeMarker,
    PropertyMetadataSet,
)
from .interpretation import Interpretation

# By default, data gathered from this interpretation is lower cased when a string.
DEFAULT_NORMALIZATION_ARGUMENTS = {LowercaseStrings.argument_flag(): True}


class SourceNodeInterpretation(Interpretation, alias="source_node"):
    """Stores information regarding the source node.

    Within a Pipeline File, a simple usage may look like this:

    ```yaml
    interpretations:
      # Conventionally, this should be first.
      - type: source_node
        node_type: TheTypeOfTheNodeYouAreIngesting
        key:
          id: !!python/jmespath value_for_id_field
    ```

    However, more complex usages may look like this to populate a rich node.

    ```yaml
    interpretations:
      - type: source_node
        node_type: Person
        key:
          first_name: !jmespath first_name
          last_name: !jmespath last_name
        properties:
           bio: !jmespath biography
    ```

    You may also apply additional labels and field level indexes:

    ```yaml
    interpretations:
      - type: source_node
        node_type: Person
        key:
          first_name: !jmespath first_name
          last_name: !jmespath last_name
        properties:
           phone_number: !jmespath phone
        additionally_index:
            - phone_number
        additional_types:
            - Customer
    ```

    You are also allowed by default to send normalization flags in:

    ```yaml
    interpretations:
      - type: source_node
        node_type: DomainName
        key:
          name: !jmespath name
        normalization:
          drop_trailing_dots: true
    ```
    """

    __slots__ = (
        "node_type",
        "key",
        "properties",
        "additional_indexes",
        "additional_types",
        "norm_args",
    )

    def __init__(
        self,
        node_type: StaticValueOrValueProvider,
        key: Dict[str, StaticValueOrValueProvider],
        properties: Optional[Dict[str, StaticValueOrValueProvider]] = None,
        additional_indexes: Optional[List[str]] = None,
        additional_types: Optional[List[str]] = None,
        normalization: Optional[Dict[str, Any]] = None,
    ):
        self.node_type = ValueProvider.guarantee_value_provider(node_type)
        self.key = ValueProvider.guarantee_provider_dictionary(key)
        self.properties = ValueProvider.guarantee_provider_dictionary(properties or {})
        self.additional_indexes = additional_indexes or []
        self.additional_types = tuple(additional_types or [])
        self.norm_args = {**DEFAULT_NORMALIZATION_ARGUMENTS, **(normalization or {})}

    def interpret(self, context: ProviderContext):
        source = context.desired_ingest.source
        source.type = self.node_type.single_value(context)
        source.key_values.apply_providers(context, self.key, self.norm_args)
        source.properties.apply_providers(context, self.properties, self.norm_args)
        source.additional_types = self.additional_types

    def gather_used_indexes(self) -> Iterable[Union[KeyIndex, FieldIndex]]:
        # NOTE: We cannot generate indexes when we do not know the type until runtime.
        if not self.node_type.is_static:
            return

        node_type = self.node_type.value
        yield KeyIndex(node_type, frozenset(self.key.keys()))
        yield FieldIndex.for_ttl_timestamp(node_type)
        for field in self.additional_indexes:
            yield FieldIndex(node_type, field, object_type=GraphObjectType.NODE)

    def gather_object_shapes(self) -> Iterable[GraphObjectShape]:
        # NOTE: We cannot generate schemas when we do not know the type until runtime.
        if not self.node_type.is_static:
            return

        node_type = self.node_type.value
        yield GraphObjectShape(
            graph_object_type=GraphObjectType.NODE,
            object_type=KnownTypeMarker.fulfilling_source_node(node_type),
            properties=PropertyMetadataSet.from_names(
                self.key.keys(), self.properties.keys()
            ),
        )



from typing import Iterable

from ..model import IngestionHook, Node, RelationshipWithNodes, TimeToLiveConfiguration
from ..pipeline.meta import get_context
from ..schema.indexes import FieldIndex, KeyIndex
from .query_executor import (
    OperationOnNodeIdentity,
    OperationOnRelationshipIdentity,
    QueryExecutor,
)

NODE_STAT = "Node Upsert Operations"
RELATIONSHIP_STAT = "Relationship Upsert Operations"
KEY_INDEX_STAT = "Key Index Upsert Operations"
FIELD_INDEX_STAT = "Field Index Upsert Operations"
TTL_STAT = "Time to Live Operations"
HOOK_STAT = "Ingest Hooks Executed"


class QueryExecutorWithStatistics(QueryExecutor):
    __slots__ = ("inner",)

    def __init__(self, inner: QueryExecutor) -> None:
        self.inner = inner

    async def upsert_nodes_in_bulk_with_same_operation(
        self, operation: OperationOnNodeIdentity, nodes: Iterable[Node]
    ):
        await self.inner.upsert_nodes_in_bulk_with_same_operation(operation, nodes)
        get_context().increment_stat(NODE_STAT, len(nodes))

    async def upsert_relationships_in_bulk_of_same_operation(
        self,
        shape: OperationOnRelationshipIdentity,
        relationships: Iterable[RelationshipWithNodes],
    ):
        await self.inner.upsert_relationships_in_bulk_of_same_operation(
            shape, relationships
        )
        get_context().increment_stat(RELATIONSHIP_STAT, len(relationships))

    async def upsert_key_index(self, index: KeyIndex):
        await self.inner.upsert_key_index(index)
        get_context().increment_stat(KEY_INDEX_STAT)

    async def upsert_field_index(self, index: FieldIndex):
        await self.inner.upsert_field_index(index)
        get_context().increment_stat(FIELD_INDEX_STAT)

    async def perform_ttl_op(self, config: TimeToLiveConfiguration):
        await self.inner.perform_ttl_op(config)
        get_context().increment_stat(TTL_STAT)

    async def execute_hook(self, hook: IngestionHook):
        await self.inner.execute_hook(hook)
        get_context().increment_stat(HOOK_STAT)



from dataclasses import dataclass
from typing import Any, Dict

from ..file_io import LazyLoadedArgument


@dataclass(slots=True, frozen=True)
class Target:
    name: str
    connector_config: Dict[str, Any]

    @property
    def resolved_connector_config(self):
        return LazyLoadedArgument.resolve_if_needed(self.connector_config)

    @property
    def connector(self):
        from ..databases import DatabaseConnector

        return DatabaseConnector.from_database_args(**self.resolved_connector_config)

    def make_writer(self, **writer_args):
        from ..databases import GraphDatabaseWriter

        return GraphDatabaseWriter.from_file_data(
            **writer_args, **self.resolved_connector_config
        )

    def make_type_retriever(self):
        return self.connector.make_type_retriever()

    def to_file_data(self):
        return self.connector_config



from dataclasses import dataclass
from typing import Set

from ..file_io import LoadsFromYamlFile
from ..pipeline.scope_config import ScopeConfig


@dataclass
class PluginConfiguration(LoadsFromYamlFile):
    """A `PluginScope` represents a collection of configuration for a plugin.

    A config is a key value pair object in a nodestream scope including plugins.
    It contains a collection of configuration key value pairs to be used by the pipelines of a scope.
    """

    name: str
    config: ScopeConfig = None
    targets: Set[str] = None

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                "name": str,
                Optional("config"): ScopeConfig.describe_yaml_schema(),
                Optional("targets"): [str],
            }
        )

    @classmethod
    def from_file_data(cls, data) -> "PluginConfiguration":
        name = data.pop("name")
        targets = data.pop("targets", [])
        config = data.pop("config")
        return cls(name, ScopeConfig.from_file_data(config), targets)

    def get_config_value(self, key):
        return self.config.get_config_value(key)



import os
from dataclasses import dataclass, field
from logging import Logger, getLogger
from typing import Any, Callable

from psutil import Process

from .meta import PipelineContext


def no_op(*_, **__):
    """A no-op function that does nothing."""
    pass


def get_max_mem_mb():
    """Get the maximum memory used by the current process in MB.

    Returns:
        int: The maximum memory used by the current process in MB.
    """
    process = Process(os.getpid())
    return process.memory_info().rss


@dataclass
class PipelineProgressReporter:
    """A `PipelineProgressReporter` is a utility that can be used to report on the progress of a pipeline."""

    reporting_frequency: int = 1000
    logger: Logger = field(default_factory=getLogger)
    callback: Callable[[int, Any], None] = field(default=no_op)
    on_start_callback: Callable[[], None] = field(default=no_op)
    on_finish_callback: Callable[[PipelineContext], None] = field(default=no_op)

    @classmethod
    def for_testing(cls, results_list: list) -> "PipelineProgressReporter":
        """Create a `PipelineProgressReporter` for testing.

        This method is intended to be used for testing purposes only. It will create a
        `PipelineProgressReporter` with the default values for testing.

        Args:
            results_list: The list to append results to.

        Returns:
            PipelineProgressReporter: A `PipelineProgressReporter` for testing.
        """
        return cls(
            reporting_frequency=1,
            logger=getLogger("test"),
            callback=lambda _, record: results_list.append(record),
            on_start_callback=no_op,
            on_finish_callback=no_op,
        )

    def report(self, index, record):
        if index % self.reporting_frequency == 0:
            self.logger.info(
                "Records Processed",
                extra={"index": index, "max_memory": get_max_mem_mb()},
            )
            self.callback(index, record)



from typing import Any, Dict, Iterable, Type

from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import StaticValueOrValueProvider, ValueProvider


class StringFormattingValueProvider(ValueProvider):
    """A `ValueProvider` that uses string formatting to produce a value."""

    __slots__ = ("fmt", "subs")

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!format", lambda loader, node: cls(**loader.construct_mapping(node))
        )

    def __init__(
        self,
        fmt: StaticValueOrValueProvider,
        **subs: Dict[str, StaticValueOrValueProvider],
    ) -> None:
        self.fmt = ValueProvider.guarantee_value_provider(fmt)
        self.subs = ValueProvider.guarantee_provider_dictionary(subs)

    def single_value(self, context: ProviderContext) -> Any:
        if (fmt := self.fmt.single_value(context)) is None:
            return None

        subs = {
            field: provider.single_value(context)
            for field, provider in self.subs.items()
        }

        return fmt.format(**subs)

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        value = self.single_value(context)
        return [value] if value else []


SafeDumper.add_representer(
    StringFormattingValueProvider,
    lambda dumper, mapping: dumper.represent_mapping(
        "!format", dict(fmt=mapping.fmt, **mapping.subs)
    ),
)



from typing import Type

from ...project import Project
from ...project.audits import Audit, AuditPrinter
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

DEFAULT_SCOPE_NAME = "default"


class CommandAuditPrinter(AuditPrinter):
    def __init__(self, command: NodestreamCommand):
        self.command = command

    def print_success(self, message: str) -> None:
        self.command.line(f"<info>{message}</info>")

    def print_failure(self, message: str) -> None:
        self.command.line(f"<error>{message}</error>")

    def print_warning(self, message: str) -> None:
        self.command.line(f"<comment>{message}</comment>")


class RunAudit(Operation):
    def __init__(self, project: Project, audit_cls: Type[Audit]) -> None:
        self.project = project
        self.audit_cls = audit_cls

    def make_audit(self, command: NodestreamCommand) -> Audit:
        return self.audit_cls(CommandAuditPrinter(command))

    async def perform(self, command: NodestreamCommand):
        audit = self.make_audit(command)
        await audit.run(self.project)
        return audit.failure_count



from typing import Any, Dict

from .creation_rules import NodeCreationRule, RelationshipCreationRule
from .desired_ingestion import DesiredIngestion, RelationshipWithNodes
from .graph_objects import (
    Node,
    NodeIdentityShape,
    PropertySet,
    Relationship,
    RelationshipIdentityShape,
)
from .ingestion_hooks import IngestionHook, IngestionHookRunRequest
from .ttl import TimeToLiveConfiguration

JsonLikeDocument = Dict[str, Any]


__all__ = (
    "DesiredIngestion",
    "RelationshipWithNodes",
    "Node",
    "NodeIdentityShape",
    "PropertySet",
    "Relationship",
    "RelationshipIdentityShape",
    "NodeCreationRule",
    "IngestionHook",
    "IngestionHookRunRequest",
    "TimeToLiveConfiguration",
    "JsonLikeDocument",
    "RelationshipCreationRule",
)



from abc import ABC
from typing import Type

from yaml import SafeLoader

from ...pluggable import Pluggable
from ...subclass_registry import SubclassRegistry

ARGUMENT_RESOLVER_REGISTRY = SubclassRegistry()


@ARGUMENT_RESOLVER_REGISTRY.connect_baseclass
class ArgumentResolver(Pluggable, ABC):
    """An `ArgumentResolver` is a class that can resolve a value by injecting into the yaml parser."""

    entrypoint_name = "argument_resolvers"

    @staticmethod
    def resolve_argument(value):
        return value

    @classmethod
    def resolve_argument_with_alias(cls, tag, value):
        ArgumentResolver.import_all()
        resolver = ARGUMENT_RESOLVER_REGISTRY.get(tag)
        return resolver.resolve_argument(value)

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        def construct_argument(loader, node):
            return cls.resolve_argument(loader.construct_scalar(node))

        yaml_tag = f"!{ARGUMENT_RESOLVER_REGISTRY.name_for(cls)}"
        loader.add_constructor(yaml_tag, construct_argument)



from ...schema.indexes import FieldIndex, KeyIndex
from ...schema.schema import GraphObjectType
from .query import Query

KEY_INDEX_QUERY_FORMAT = "CREATE CONSTRAINT {constraint_name} IF NOT EXISTS FOR (n:`{type}`) REQUIRE ({key_pattern}) IS UNIQUE"
ENTERPRISE_KEY_INDEX_QUERY_FORMAT = "CREATE CONSTRAINT {constraint_name} IF NOT EXISTS FOR (n:`{type}`) REQUIRE ({key_pattern}) IS NODE KEY"
NODE_FIELD_INDEX_QUERY_FORMAT = (
    "CREATE INDEX {constraint_name} IF NOT EXISTS FOR (n:`{type}`) ON (n.`{field}`)"
)
REL_FIELD_INDEX_QUERY_FORMAT = "CREATE INDEX {constraint_name} IF NOT EXISTS FOR ()-[r:`{type}`]-() ON (r.`{field}`)"


def key_index_from_format(key_index: KeyIndex, format: str) -> Query:
    key_pattern = ",".join(f"n.`{p}`" for p in sorted(key_index.identity_keys))
    constraint_name = f"{key_index.type}_node_key"
    statement = format.format(
        constraint_name=constraint_name,
        key_pattern=key_pattern,
        type=key_index.type,
    )
    return Query.from_statement(statement)


class Neo4jIndexQueryBuilder:
    """Creates index creation queries that will work for any Neo4j Database Supported."""

    def create_key_index_query(self, key_index: KeyIndex) -> Query:
        """Creates a key index using a "Unique node property constraint" constraint which is supported by community.

        see: https://neo4j.com/docs/cypher-manual/current/constraints/
        """
        return key_index_from_format(key_index, KEY_INDEX_QUERY_FORMAT)

    def create_field_index_query(self, field_index: FieldIndex) -> Query:
        """Creates a filed index using a 'Range Index' or equivalent.

        see: https://neo4j.com/docs/cypher-manual/current/indexes-for-search-performance/#indexes-create-indexes
        """
        constraint_name = f"{field_index.type}_{field_index.field}_additional_index"
        format = (
            REL_FIELD_INDEX_QUERY_FORMAT
            if field_index.object_type == GraphObjectType.RELATIONSHIP
            else NODE_FIELD_INDEX_QUERY_FORMAT
        )
        statement = format.format(
            constraint_name=constraint_name,
            type=field_index.type,
            field=field_index.field,
        )
        return Query.from_statement(statement)


class Neo4jEnterpriseIndexQueryBuilder(Neo4jIndexQueryBuilder):
    """Creates index creation query that will work only for Neo4j Enterprise"""

    def create_key_index_query(self, key_index: KeyIndex) -> Query:
        """Generates a key index using a `Node key constraint` which is an enterprise feature.

        see: https://neo4j.com/docs/cypher-manual/current/constraints/
        """
        return key_index_from_format(key_index, ENTERPRISE_KEY_INDEX_QUERY_FORMAT)



from typing import Any, Iterable, Type

from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import StaticValueOrValueProvider, ValueProvider


class MappingValueProvider(ValueProvider):
    """A `ValueProvider` that uses a mapping to extract values from a document."""

    __slots__ = ("mapping_name", "key")

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!mapping",
            lambda loader, node: MappingValueProvider(**loader.construct_mapping(node)),
        )

    def __init__(self, mapping_name: str, key: StaticValueOrValueProvider) -> None:
        self.mapping_name = mapping_name
        self.key = ValueProvider.guarantee_value_provider(key)

    def single_value(self, context: ProviderContext) -> Any:
        mapping = context.mappings.get(self.mapping_name)
        if not mapping:
            return

        key = self.key.single_value(context)
        if not key:
            return

        value = ValueProvider.guarantee_value_provider(mapping.get(key))
        return value.single_value(context)

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        value = self.single_value(context)
        return [value] if value else []


SafeDumper.add_representer(
    MappingValueProvider,
    lambda dumper, mapping: dumper.represent_mapping(
        "!mapping", {"mapping_name": mapping.mapping_name, "key": mapping.key}
    ),
)



from dataclasses import dataclass
from typing import Dict

from ..file_io import LazyLoadedArgument, LoadsFromYamlFile


@dataclass
class ScopeConfig(LoadsFromYamlFile):
    """A `ScopeConfig` represents a collection of configuration for a scope.

    A config is a key value pair object in a nodestream scope including plugins.
    It contains a collection of configuration key value pairs to be used by the pipelines of a scope.
    """

    config: Dict[str, any]

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Schema

        return Schema(dict)

    @classmethod
    def from_file_data(cls, data) -> "ScopeConfig":
        """Creates a config object from file data.

        The file data should be a dictionary mapping of names with a dictionary of config key value pairs and
        should be validated by the schema returned by `describe_yaml_schema`.

        This method should not be called directly. Instead, use `LoadsFromYamlFile.load_from_yaml_file`.

        Args:
            data (Dict): The file data to create the config from.

        Returns:
            Config: The config created from the file data.
        """

        return cls(data)

    def get_config_value(self, key):
        return LazyLoadedArgument.resolve_if_needed(self.config.get(key))



import asyncio
from logging import getLogger
from typing import Any, Iterable, List, Optional

from confluent_kafka import Consumer, KafkaException

from .extractor import StreamConnector

DEFAULT_GROUP_ID = "nodestream"


class KafkaConsumerFactory:
    @classmethod
    def build_consumer_config(
        cls,
        offset_reset: str,
        security_protocol: str,
        group_id: str,
        bootstrap_servers: List[str],
    ):
        return {
            "bootstrap.servers": ",".join(bootstrap_servers),
            "auto.offset.reset": offset_reset,
            "security.protocol": security_protocol,
            "group.id": group_id,
        }

    @classmethod
    def new(
        cls,
        topic_name: str,
        group_id: str,
        bootstrap_servers: List[str],
        offset_reset: str,
        security_protocol: str,
    ):
        consumer = Consumer(
            cls.build_consumer_config(
                offset_reset, security_protocol, group_id, bootstrap_servers
            )
        )
        consumer.subscribe([topic_name])
        return consumer


class KafkaStreamConnector(StreamConnector, alias="kafka"):
    """A KafkaStreamConnector implements the StreamConnector interface for Kafka.

    The KafkaStreamConnector uses the aiokafka library to connect to a
    Kafka cluster and poll data from it.
    """

    def __init__(
        self,
        bootstrap_servers: List[str],
        topic: str,
        group_id: Optional[str] = None,
        offset_reset: str = "latest",
        security_protocol: str = "PLAINTEXT",
        max_records: int = 10,
        poll_timeout: int = 30,
    ):
        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.max_records = 10
        self.group_id = group_id or DEFAULT_GROUP_ID
        self.consumer: Consumer = None
        self.offset_reset = offset_reset
        self.security_protocol = security_protocol
        self.max_records = max_records
        self.poll_timeout = poll_timeout
        self.logger = getLogger(__name__)

    async def connect(self):
        self.logger.debug("Starting Connection to Kafka Topic %s", self.topic)
        self.consumer = KafkaConsumerFactory.new(
            topic_name=self.topic,
            group_id=self.group_id,
            bootstrap_servers=self.bootstrap_servers,
            offset_reset=self.offset_reset,
            security_protocol=self.security_protocol,
        )
        self.logger.info("Connected to Kafka Topic %s", self.topic)

    async def disconnect(self):
        self.consumer.close()

    async def poll(self) -> Iterable[Any]:
        results = []
        for _ in range(self.max_records):
            try:
                result = await self.get_next_messsage()
                if result is None:
                    self.logger.debug("Polling returned no messages")
                    continue
                results.append(result)
            except Exception:
                self.logger.exception("error while polling Kafka messages")
                break
        return results

    async def get_next_messsage(self):
        loop = asyncio.get_running_loop()
        msg = await loop.run_in_executor(None, self.consumer.poll, self.poll_timeout)
        if msg is None:
            return None
        message_value = self.process_message(msg)
        return message_value

    def process_message(self, msg):
        if msg.error():
            raise KafkaException(msg.error())
        else:
            return msg.value()



from typing import Iterable

from yaml import safe_dump

from ...pipeline import PipelineInitializationArguments, PipelineProgressReporter
from ...pipeline.meta import PipelineContext
from ...project import Project, RunRequest
from ...project.pipeline_definition import PipelineDefinition
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

STATS_TABLE_COLS = ["Statistic", "Value"]

ERROR_NO_PIPELINES_FOUND = "<error>No pipelines with the provided name were found in your project. If you didn't provide a name, you have no pipelines.</error>"
HINT_CHECK_PIPELINE_NAME = "<info>HINT: Check that the pipelines you are trying to run are named correctly and in the registry.</info>"
HINT_USE_NODESTREAM_SHOW = "<info>HINT: You can view your project's pipelines by running 'nodestream show'. </info>"


class RunPipeline(Operation):
    def __init__(self, project: Project) -> None:
        self.project = project

    def get_pipelines_to_run(
        self, command: NodestreamCommand
    ) -> Iterable[PipelineDefinition]:
        supplied_commands = command.argument("pipelines")
        if supplied_commands:
            return [
                self.project.get_pipeline_by_name(name) for name in supplied_commands
            ]
        return self.project.get_all_pipelines()

    async def perform(self, command: NodestreamCommand):
        pipelines_ran = 0
        for pipeline in self.get_pipelines_to_run(command):
            request = self.make_run_request(command, pipeline)
            pipelines_ran += await self.project.run(request)

        if pipelines_ran == 0:
            command.line(ERROR_NO_PIPELINES_FOUND)
            command.line(HINT_CHECK_PIPELINE_NAME)
            command.line(HINT_USE_NODESTREAM_SHOW)

    def get_writer_steps_for_specified_targets(
        self, command: NodestreamCommand, pipeline: PipelineDefinition
    ):
        targets_names = self.combine_targets_from_command_and_pipeline(
            command, pipeline
        )
        for target_name in targets_names:
            target = self.project.get_target_by_name(target_name)
            if target:
                yield target.make_writer()
            else:
                command.line(
                    f"<error>Target '{target_name}' not found in project. Ignoring.</error>"
                )

    def combine_targets_from_command_and_pipeline(
        self, command: NodestreamCommand, pipeline: PipelineDefinition
    ):
        from_cli = set(command.option("target") or {})
        from_pipeline = set(pipeline.targets or {})
        return from_cli.union(from_pipeline)

    def make_run_request(
        self, command: NodestreamCommand, pipeline: PipelineDefinition
    ) -> RunRequest:
        def print_effective_config(config):
            if command.is_very_verbose:
                command.line("<info>Effective configuration:</info>")
                command.line(f"<info>{safe_dump(config)}</info>")

        return RunRequest(
            pipeline_name=pipeline.name,
            initialization_arguments=PipelineInitializationArguments(
                annotations=command.option("annotations"),
                step_outbox_size=int(command.option("step-outbox-size")),
                on_effective_configuration_resolved=print_effective_config,
                extra_steps=list(
                    self.get_writer_steps_for_specified_targets(command, pipeline)
                ),
            ),
            progress_reporter=self.create_progress_reporter(command, pipeline.name),
        )

    def get_progress_indicator(
        self, command: NodestreamCommand, pipeline_name: str
    ) -> "ProgressIndicator":
        if command.has_json_logging_set:
            return ProgressIndicator(command, pipeline_name)

        return SpinnerProgressIndicator(command, pipeline_name)

    def create_progress_reporter(
        self, command: NodestreamCommand, pipeline_name: str
    ) -> PipelineProgressReporter:
        indicator = self.get_progress_indicator(command, pipeline_name)
        return PipelineProgressReporter(
            reporting_frequency=int(command.option("reporting-frequency")),
            callback=indicator.progress_callback,
            on_start_callback=indicator.on_start,
            on_finish_callback=indicator.on_finish,
        )


class ProgressIndicator:
    def __init__(self, command: NodestreamCommand, pipeline_name: str) -> None:
        self.command = command
        self.pipeline_name = pipeline_name

    def on_start(self):
        pass

    def progress_callback(self, _, __):
        pass

    def on_finish(self, context: PipelineContext):
        pass


class SpinnerProgressIndicator(ProgressIndicator):
    def on_start(self):
        self.progress = self.command.progress_indicator()
        self.progress.start(f"Running pipeline: '{self.pipeline_name}'")

    def progress_callback(self, index, _):
        self.progress.set_message(
            f"Currently processing record at index: <info>{index}</info>"
        )

    def on_finish(self, context: PipelineContext):
        self.progress.finish(f"Finished running pipeline: '{self.pipeline_name}'")

        stats = ((k, str(v)) for k, v in context.stats.items())
        table = self.command.table(STATS_TABLE_COLS, stats)
        table.render()



from typing import Any, Dict

from ...pipeline.value_providers import ProviderContext, ValueProvider
from .transformer import Transformer


class ValueProjection(Transformer):
    def __init__(
        self,
        projection: ValueProvider,
        additional_values: Dict[str, ValueProvider] = {},
    ) -> None:
        self.projection = projection
        self.additional_values = additional_values

    def fetch_additional_values(self, context):
        return {
            key: provider.single_value(context)
            for key, provider in self.additional_values.items()
        }

    async def transform_record(self, record: Any):
        context = ProviderContext.fresh(record)
        for result in self.projection.many_values(context):
            yield dict(**self.fetch_additional_values(context), **result)



from typing import Type

from ...project.audits import Audit
from ..operations import InitializeProject, RunAudit
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION


class AuditCommand(NodestreamCommand):
    audit: Type[Audit] = Audit
    options = [PROJECT_FILE_OPTION]

    async def handle_async(self):
        project = await self.run_operation(InitializeProject())
        return await self.run_operation(RunAudit(project, self.audit))

    @classmethod
    def for_audit(cls, audit_cls: Type[Audit]) -> Type["AuditCommand"]:
        class NewAuditCommand(AuditCommand):
            audit = audit_cls
            name = f"audit {audit_cls.name}"
            description = audit_cls.description

        return NewAuditCommand



import re
from datetime import datetime, timedelta
from functools import cache, wraps
from typing import Iterable

from cymple.builder import NodeAfterMergeAvailable, NodeAvailable, QueryBuilder

from ...model import (
    Node,
    NodeCreationRule,
    Relationship,
    RelationshipCreationRule,
    RelationshipIdentityShape,
    RelationshipWithNodes,
    TimeToLiveConfiguration,
)
from ...schema.schema import GraphObjectType
from ..query_executor import OperationOnNodeIdentity, OperationOnRelationshipIdentity
from .query import Query, QueryBatch

PROPERTIES_PARAM_NAME = "properties"
ADDITIONAL_LABELS_PARAM_NAME = "additional_labels"
GENERIC_NODE_REF_NAME = "node"
FROM_NODE_REF_NAME = "from_node"
TO_NODE_REF_NAME = "to_node"
RELATIONSHIP_REF_NAME = "rel"
PARAMETER_CORRECTION_REGEX = re.compile(r"\"(params.__\w+)\"")
DELETE_NODE_QUERY = "MATCH (n) WHERE id(n) = id DETACH DELETE n"
DELETE_REL_QUERY = "MATCH ()-[r]->() WHERE id(r) = id DELETE r"


def correct_parameters(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        query = f(*args, **kwargs)
        return PARAMETER_CORRECTION_REGEX.sub(r"\1", query)

    return wrapper


def generate_prefixed_param_name(property_name: str, prefix: str) -> str:
    return f"__{prefix}_{property_name}"


def generate_properties_set_with_prefix(properties: Iterable[str], prefix: str):
    return {
        prop: f"params.{generate_prefixed_param_name(prop, prefix)}"
        for prop in properties
    }


def generate_where_set_with_prefix(properties: frozenset, prefix: str):
    return {
        f"{prefix}.{prop}": f"params.{generate_prefixed_param_name(prop, prefix)}"
        for prop in properties
    }


@cache
def _match_node(
    node_operation: OperationOnNodeIdentity, name=GENERIC_NODE_REF_NAME
) -> NodeAvailable:
    op = "=~" if node_operation.node_creation_rule == NodeCreationRule.FUZZY else "="
    identity = node_operation.node_identity
    props = generate_where_set_with_prefix(identity.keys, name)
    return (
        QueryBuilder()
        .match()
        .node(labels=identity.type, ref_name=name)
        .where_multiple(props, comparison_operator=op)
    )


@cache
def _merge_node(
    node_operation: OperationOnNodeIdentity, name=GENERIC_NODE_REF_NAME
) -> NodeAfterMergeAvailable:
    properties = generate_properties_set_with_prefix(
        node_operation.node_identity.keys, name
    )
    return (
        QueryBuilder()
        .merge()
        .node(
            labels=node_operation.node_identity.type,
            ref_name=name,
            properties=properties,
        )
    )


@cache
def _make_relationship(
    rel_identity: RelationshipIdentityShape, creation_rule: RelationshipCreationRule
):
    keys = generate_properties_set_with_prefix(rel_identity.keys, RELATIONSHIP_REF_NAME)
    match_rel_query = (
        QueryBuilder()
        .match_optional()
        .node(ref_name=FROM_NODE_REF_NAME)
        .related_to(
            ref_name=RELATIONSHIP_REF_NAME,
            properties=keys,
            label=rel_identity.type,
        )
        .node(ref_name=TO_NODE_REF_NAME)
    )

    create_rel_query = str(match_rel_query).replace("OPTIONAL MATCH ", "CREATE")
    set_properties_query = f"SET {RELATIONSHIP_REF_NAME} += params.{generate_prefixed_param_name(PROPERTIES_PARAM_NAME, RELATIONSHIP_REF_NAME)}"
    if creation_rule == RelationshipCreationRule.CREATE:
        return f"{create_rel_query} {set_properties_query}"

    return f"""
    {match_rel_query}
    FOREACH (x IN CASE WHEN {RELATIONSHIP_REF_NAME} IS NULL THEN [1] ELSE [] END |
        {create_rel_query} {set_properties_query})
    FOREACH (i in CASE WHEN {RELATIONSHIP_REF_NAME} IS NOT NULL THEN [1] ELSE [] END |
        {set_properties_query})
    """


class Neo4jIngestQueryBuilder:
    def __init__(self, apoc_iterate: bool):
        self.apoc_iterate = apoc_iterate

    @cache
    @correct_parameters
    def generate_update_node_operation_query_statement(
        self,
        operation: OperationOnNodeIdentity,
    ) -> str:
        """Generate a query to update a node in the database given a node type and a match strategy."""

        if operation.node_creation_rule == NodeCreationRule.EAGER:
            query = str(_merge_node(operation))
        else:
            query = str(_match_node(operation))

        if operation.node_identity.additional_types:
            query += f" WITH {GENERIC_NODE_REF_NAME}, params CALL apoc.create.addLabels({GENERIC_NODE_REF_NAME}, params.{generate_prefixed_param_name(ADDITIONAL_LABELS_PARAM_NAME, GENERIC_NODE_REF_NAME)}) yield node as _"

        query += f" SET {GENERIC_NODE_REF_NAME} += params.{generate_prefixed_param_name(PROPERTIES_PARAM_NAME, GENERIC_NODE_REF_NAME)}"
        return query

    def generate_update_node_operation_params(self, node: Node) -> dict:
        """Generate the parameters for a query to update a node in the database."""

        params = self.generate_node_key_params(node)
        params[
            generate_prefixed_param_name(PROPERTIES_PARAM_NAME, GENERIC_NODE_REF_NAME)
        ] = node.properties
        params[
            generate_prefixed_param_name(
                ADDITIONAL_LABELS_PARAM_NAME, GENERIC_NODE_REF_NAME
            )
        ] = node.additional_types

        return params

    def generate_node_key_params(self, node: Node, name=GENERIC_NODE_REF_NAME) -> dict:
        """Generate the parameters for a query to update a node in the database."""

        return {
            generate_prefixed_param_name(k, name): v for k, v in node.key_values.items()
        }

    @cache
    @correct_parameters
    def generate_update_relationship_operation_query_statement(
        self,
        operation: OperationOnRelationshipIdentity,
    ) -> str:
        """Generate a query to update a relationship in the database given a relationship operation."""

        match_from_node_segment = _match_node(operation.from_node, FROM_NODE_REF_NAME)
        match_to_node_segment = _match_node(operation.to_node, TO_NODE_REF_NAME)
        merge_rel_segment = _make_relationship(
            operation.relationship_identity, operation.relationship_creation_rule
        )
        return f"{match_from_node_segment} {match_to_node_segment} {merge_rel_segment}"

    def generate_update_rel_params(self, rel: Relationship) -> dict:
        """Generate the parameters for a query to update a relationship in the database."""

        params = {
            generate_prefixed_param_name(k, RELATIONSHIP_REF_NAME): v
            for k, v in rel.key_values.items()
        }
        params[
            generate_prefixed_param_name(PROPERTIES_PARAM_NAME, RELATIONSHIP_REF_NAME)
        ] = rel.properties

        return params

    def generate_update_rel_between_nodes_params(
        self, rel: RelationshipWithNodes
    ) -> dict:
        """Generate the parameters for a query to update a relationship in the database."""

        params = self.generate_update_rel_params(rel.relationship)
        params.update(self.generate_node_key_params(rel.from_node, FROM_NODE_REF_NAME))
        params.update(self.generate_node_key_params(rel.to_node, TO_NODE_REF_NAME))
        return params

    def generate_batch_update_node_operation_batch(
        self,
        operation: OperationOnNodeIdentity,
        nodes: Iterable[Node],
    ) -> QueryBatch:
        """Generate a batch of queries to update nodes in the database in the same way of the same type."""

        query_statement = self.generate_update_node_operation_query_statement(operation)
        params = [self.generate_update_node_operation_params(node) for node in nodes]
        return QueryBatch(query_statement, params)

    def generate_batch_update_relationship_query_batch(
        self,
        operation: OperationOnRelationshipIdentity,
        relationships: Iterable[RelationshipWithNodes],
    ) -> QueryBatch:
        """Generate a batch of queries to update relationships in the database in the same way of the same type."""

        query = self.generate_update_relationship_operation_query_statement(operation)
        params = [
            self.generate_update_rel_between_nodes_params(rel) for rel in relationships
        ]
        return QueryBatch(query, params)

    def generate_ttl_match_query(self, config: TimeToLiveConfiguration) -> Query:
        earliest_allowed_time = datetime.utcnow() - timedelta(
            hours=config.expiry_in_hours
        )
        params = {"earliest_allowed_time": earliest_allowed_time}
        if config.custom_query is not None:
            return Query(config.custom_query, params)

        query_builder = QueryBuilder()
        ref_name = "x"

        if config.graph_object_type == GraphObjectType.NODE:
            query_builder = query_builder.match().node(
                labels=config.object_type, ref_name=ref_name
            )
        else:
            query_builder = (
                query_builder.match()
                .node()
                .related_to(label=config.object_type, ref_name=ref_name)
                .node()
            )

        query_builder = query_builder.where_literal(
            f"{ref_name}.last_ingested_at <= $earliest_allowed_time"
        ).return_literal(f"id({ref_name}) as id")

        return Query(str(query_builder), params)

    def generate_ttl_query_from_configuration(
        self, config: TimeToLiveConfiguration
    ) -> Query:
        ttl_match_query = self.generate_ttl_match_query(config)
        operation = (
            DELETE_NODE_QUERY
            if config.graph_object_type == GraphObjectType.NODE
            else DELETE_REL_QUERY
        )
        return ttl_match_query.feed_batched_query(operation)



import os

from .argument_resolver import ArgumentResolver


class EnvironmentResolver(ArgumentResolver, alias="env"):
    """An `EnvironmentResolver` is an `ArgumentResolver` that can resolve an environment variable into its value."""

    @staticmethod
    def resolve_argument(variable_name):
        return os.environ.get(variable_name)



from neo4j import AsyncDriver, AsyncGraphDatabase

from ..copy import TypeRetriever
from ..database_connector import DatabaseConnector
from ..query_executor import QueryExecutor
from .index_query_builder import (
    Neo4jEnterpriseIndexQueryBuilder,
    Neo4jIndexQueryBuilder,
)
from .ingest_query_builder import Neo4jIngestQueryBuilder


class Neo4jDatabaseConnector(DatabaseConnector, alias="neo4j"):
    @classmethod
    def from_file_data(
        cls,
        uri: str,
        username: str,
        password: str,
        database_name: str = "neo4j",
        use_enterprise_features: bool = False,
        use_apoc: bool = True,
    ):
        driver = AsyncGraphDatabase.driver(uri, auth=(username, password))
        if use_enterprise_features:
            index_query_builder = Neo4jEnterpriseIndexQueryBuilder()
        else:
            index_query_builder = Neo4jIndexQueryBuilder()
        return cls(
            driver=driver,
            index_query_builder=index_query_builder,
            ingest_query_builder=Neo4jIngestQueryBuilder(use_apoc),
            database_name=database_name,
        )

    def __init__(
        self,
        driver: AsyncDriver,
        index_query_builder: Neo4jIndexQueryBuilder,
        ingest_query_builder: Neo4jIngestQueryBuilder,
        database_name: str,
    ) -> None:
        self.driver = driver
        self.index_query_builder = index_query_builder
        self.ingest_query_builder = ingest_query_builder
        self.database_name = database_name

    def make_query_executor(self) -> QueryExecutor:
        from .query_executor import Neo4jQueryExecutor

        return Neo4jQueryExecutor(
            self.driver,
            self.ingest_query_builder,
            self.index_query_builder,
            self.database_name,
        )

    def make_type_retriever(self) -> TypeRetriever:
        from .type_retriever import Neo4jTypeRetriever

        return Neo4jTypeRetriever(self)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ..subclass_registry import SubclassRegistry

if TYPE_CHECKING:
    from ..model import (
        IngestionHookRunRequest,
        Node,
        RelationshipWithNodes,
        TimeToLiveConfiguration,
    )
    from ..schema.indexes import FieldIndex, KeyIndex


INGESTION_STRATEGY_REGISTRY = SubclassRegistry()


@INGESTION_STRATEGY_REGISTRY.connect_baseclass
class IngestionStrategy(ABC):
    """An IngestionStrategy represents the methods taken to commit data to a Graph Database.

    Within nodestream, many components gather and intend to commit changes to a Graph Database.
    How that data is committed and how it is stored internally is decoupled through the `IngestionStrategy` interface.

    Generally, your usage of nodestream is decoupled from `IngestionStrategy` unless you intend to provide an implementation
    of your own database writer. The writer API will give you a `DesiredIngestion` or other `Ingestible` object that needs a
    instance of an `IngestionStrategy` to apply operations to.
    """

    @abstractmethod
    async def ingest_source_node(self, source: "Node"):
        """Given a provided instance of `Node`, ensure that it is committed to the GraphDatabase."""
        raise NotImplementedError

    @abstractmethod
    async def ingest_relationship(self, relationship: "RelationshipWithNodes"):
        """Given a provided instance of `SourceNode`, ensure that the provided `Relationship` is committed to the database."""
        raise NotImplementedError

    @abstractmethod
    async def run_hook(self, request: "IngestionHookRunRequest"):
        """Runs the provided request for an IngestHook given the context."""
        raise NotImplementedError

    @abstractmethod
    async def upsert_key_index(self, index: "KeyIndex"):
        """Create a Key Index Immediately for a given Node Type.

        See information on `KeyIndex` for more information.
        """
        raise NotImplementedError

    @abstractmethod
    async def upsert_field_index(self, index: "FieldIndex"):
        """Create a Key Index Immediately for a Object Type.

        See information on `FieldIndex` for more information.
        """
        raise NotImplementedError

    @abstractmethod
    async def perform_ttl_operation(self, config: "TimeToLiveConfiguration"):
        """Perform a TTL Operation.

        See Information on `TimeToLiveConfiguration` for more Information.
        """
        raise NotImplementedError

    async def flush(self):
        """Flush any pending operations to the database."""
        pass



from .audit_command import AuditCommand
from .copy import Copy
from .new import New
from .nodestream_command import NodestreamCommand
from .print_schema import PrintSchema
from .remove import Remove
from .run import Run
from .scaffold import Scaffold
from .show import Show

__all__ = (
    "AuditCommand",
    "Copy",
    "New",
    "NodestreamCommand",
    "PrintSchema",
    "Remove",
    "Run",
    "Scaffold",
    "Show",
)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Iterable, List, Optional

import yaml

SOURCE_ALIAS = "__SOURCE_NODE__"


class GraphObjectType(str, Enum):
    RELATIONSHIP = "RELATIONSHIP"
    NODE = "NODE"

    def __str__(self) -> str:
        return self.value


class Cardinality(str, Enum):
    SINGLE = "SINGLE"
    MANY = "MANY"


class PropertyType(str, Enum):
    STRING = "STRING"
    BOOLEAN = "BOOLEAN"
    DATETIME = "DATETIME"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"


@dataclass(slots=True)
class PropertyMetadata:
    name: str
    type: PropertyType = PropertyType.STRING

    def __str__(self) -> str:
        return f"{self.name}: {self.type.value}"


@dataclass(slots=True)
class PropertyMetadataSet:
    properties: Dict[str, PropertyMetadata]

    def update(self, other: "PropertyMetadataSet"):
        for property_name, property_meta in other.properties.items():
            self.properties[property_name] = property_meta

    @classmethod
    def from_names(
        cls, *name_iterables: Iterable[str], include_timestamps: bool = True
    ):
        properties = {
            name: PropertyMetadata(name=name)
            for name_iterable in name_iterables
            for name in name_iterable
        }

        if include_timestamps:
            timestamp_name = "last_ingested_at"
            properties[timestamp_name] = PropertyMetadata(
                name=timestamp_name, type=PropertyType.DATETIME
            )

        return cls(properties=properties)

    def __str__(self) -> str:
        return "\n\t".join(str(property) for property in self.properties.values())


class TypeMarker(ABC):
    @abstractmethod
    def resolve_type(self, other_shapes: Iterable["GraphObjectShape"]) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def fulfills_alias(self, desired_alias: str) -> bool:
        raise NotImplementedError


class KnownTypeMarker(TypeMarker):
    """Represents a type that is known within the context of its creation and may be referenced as something else in the future."""

    def __init__(self, type: str, fulfills_alias: Optional[str] = None) -> None:
        self.type = type
        self.alias = fulfills_alias

    def resolve_type(
        self, other_shapes: Iterable["GraphObjectShape"]
    ) -> Optional[TypeMarker]:
        return self

    def fulfills_alias(self, desired_alias: str) -> bool:
        return self.alias == desired_alias

    def __eq__(self, o: object) -> bool:
        return isinstance(o, KnownTypeMarker) and o.type == self.type

    @classmethod
    def fulfilling_source_node(cls, type: str) -> "KnownTypeMarker":
        return cls(type=type, fulfills_alias=SOURCE_ALIAS)

    def __str__(self) -> str:
        return self.type


class UnknownTypeMarker(TypeMarker):
    """Represents a type that is not known currently and requires context from others to resolve.

    In other words, we know what role this type is, but not what the exact type is.
    """

    def __init__(self, alias: str) -> None:
        self.alias = alias

    def resolve_type(
        self, other_shapes: Iterable["GraphObjectShape"]
    ) -> Optional[TypeMarker]:
        for shape in other_shapes:
            if shape.object_type.fulfills_alias(self.alias):
                return shape.object_type

    def fulfills_alias(self, desired_alias: str) -> bool:
        return False

    def __eq__(self, o: object) -> bool:
        return isinstance(o, UnknownTypeMarker) and o.alias == self.alias

    def __str__(self) -> str:
        return f"[Uknown] Alias({self.alias}))"

    @classmethod
    def source_node(cls):
        return cls(alias=SOURCE_ALIAS)


@dataclass(slots=True)
class GraphObjectShape:
    """Defines the shape of an object in the graph.

    The shape of an object consists of three parts:

    1.) Whether its a node or a relationship.
    2.) What properties it has on it.
    3.) What the type of the node/relationship is.
    """

    graph_object_type: GraphObjectType
    object_type: TypeMarker
    properties: PropertyMetadataSet

    def overlaps_with(self, other: "GraphObjectShape") -> bool:
        return (
            self.graph_object_type == other.graph_object_type
            and self.object_type == other.object_type
        )

    def include(self, other: "GraphObjectShape"):
        if not self.overlaps_with(other):
            raise ValueError("Cannot include when the shape is not overlapping")

        self.properties.update(other.properties)

    def property_names(self):
        all_props = self.properties.properties
        return [all_props[prop].name for prop in all_props.keys()]

    def resolve_types(self, shapes: Iterable["GraphObjectShape"]):
        if object_type := self.object_type.resolve_type(shapes):
            self.object_type = object_type

    @property
    def is_node(self) -> bool:
        return self.graph_object_type == GraphObjectType.NODE

    @property
    def is_relationship(self) -> bool:
        return self.graph_object_type == GraphObjectType.RELATIONSHIP

    @property
    def has_known_type(self) -> bool:
        return isinstance(self.object_type, KnownTypeMarker)

    def __str__(self) -> str:
        return f"[{self.graph_object_type}] {self.object_type}:\n\t{self.properties}"


@dataclass(slots=True)
class PresentRelationship:
    """Indicates that a relationship of a specifc type exissts between two nodes with a certain cardinality.

    The shape of the relationship itself as well as the shapes of the node types are stored as `GraphObjectShape`.
    """

    from_object_type: TypeMarker
    to_object_type: TypeMarker
    relationship_type: TypeMarker
    from_side_cardinality: Cardinality
    to_side_cardinality: Cardinality

    def overlaps_with(self, other: "PresentRelationship") -> bool:
        return (
            self.from_object_type == other.from_object_type
            and self.to_object_type == other.to_object_type
            and self.relationship_type == other.relationship_type
        )

    def include(self, other: "PresentRelationship"):
        if not self.overlaps_with(other):
            raise ValueError(
                "Cannot include when other relationship is not overlapping"
            )

        # If multiple relationships come in, we do not have enough information to assume that
        # the relationship is anything other than MANY->MANY. For any side that is "single", we
        # do not know that the each one of the sources of the relationship will not all create a "single"
        # sided relationship off of the same nodes.
        self.from_side_cardinality = Cardinality.MANY
        self.to_side_cardinality = Cardinality.MANY

    def resolve_types(self, shapes: Iterable["GraphObjectShape"]):
        if object_type := self.from_object_type.resolve_type(shapes):
            self.from_object_type = object_type
        if object_type := self.to_object_type.resolve_type(shapes):
            self.to_object_type = object_type

        self.relationship_type = self.relationship_type.resolve_type(shapes)


@dataclass(frozen=True, slots=True)
class GraphSchema:
    """Defines the Schema of a graph by specifiying the node and relationship types and how they interconnect.

    See `GraphObjectShape` and `PresentRelationship` for more details.
    """

    object_shapes: List[GraphObjectShape]
    relationships: List[PresentRelationship]

    @classmethod
    def empty(cls):
        return cls(object_shapes=[], relationships=[])

    def merge(self, other: "GraphSchema") -> "GraphSchema":
        all_shapes = self.object_shapes + other.object_shapes
        all_rels = self.relationships + other.relationships
        return GraphSchema(
            object_shapes=_merge_overlapping_items(all_shapes, all_shapes),
            relationships=_merge_overlapping_items(all_rels, all_shapes),
        )

    def known_node_types(self) -> Iterable[GraphObjectShape]:
        for shape in self.object_shapes:
            if shape.is_node and shape.has_known_type:
                yield shape

    def known_relationship_types(self) -> Iterable[GraphObjectShape]:
        for shape in self.object_shapes:
            if shape.is_relationship and shape.has_known_type:
                yield shape

    def apply_type_overrides_from_file(self, file_path: Path):
        overrides = GraphSchemaOverrides.from_file(file_path)
        self.apply_overrides(overrides)

    def apply_overrides(self, overrides: "GraphSchemaOverrides"):
        overrides.apply_to(self)


@dataclass(frozen=True, slots=True)
class PropertyOverride:
    type: Optional[PropertyType]

    @classmethod
    def from_file_data(cls, file_data) -> "GraphSchemaOverrides":
        type_str = file_data.get("type")
        type = None if type_str is None else PropertyType(type_str)
        return cls(type=type)

    def apply_to(self, property_metadata: PropertyMetadata):
        if self.type:
            property_metadata.type = self.type


@dataclass(frozen=True, slots=True)
class PropertyOverrides:
    properties: Dict[str, PropertyOverride]

    @classmethod
    def from_file_data(cls, file_data) -> "GraphSchemaOverrides":
        properties = {
            key: PropertyOverride.from_file_data(value)
            for key, value in file_data.items()
        }
        return cls(properties=properties)

    def apply_to(self, graph_object_shape: GraphObjectShape):
        for property_name, override in self.properties.items():
            if property_name in graph_object_shape.properties.properties:
                override.apply_to(
                    graph_object_shape.properties.properties[property_name]
                )


@dataclass(frozen=True, slots=True)
class GraphSchemaOverrides:
    property_overrides: Dict[str, PropertyOverrides]

    @classmethod
    def from_file(cls, file_path: Path) -> "GraphSchemaOverrides":
        with open(file_path) as f:
            overrides_data = yaml.safe_load(f)
            property_overrides = {
                key: PropertyOverrides.from_file_data(value)
                for key, value in overrides_data.get("properties", {}).items()
            }
            return cls(property_overrides=property_overrides)

    def apply_to(self, schema: GraphSchema):
        for type_name, overrides in self.property_overrides.items():
            for shape in schema.known_node_types():
                if shape.object_type.type == type_name:
                    overrides.apply_to(shape)


def _merge_overlapping_items(unmerged_items, shapes):
    """Merges items that refer to the same inner types together and resolves ambiguities in types that are not known between components."""

    distinct_items = []

    for new_item in unmerged_items:
        new_item.resolve_types(shapes)
        for current_item in distinct_items:
            if new_item.overlaps_with(current_item):
                current_item.include(new_item)
                break
        else:
            distinct_items.append(new_item)

    return distinct_items


class IntrospectiveIngestionComponent(ABC):
    """An IntrospectableIngestionComponent is a componet that can be interrogated what it contributes the Graph Database Schema.

    Nearly all components in the transformation part of the ingestion stack (Interpreters, Interpretations, and Pipelines) are
    `IntrospectableIngestionComponent`s. Leaf components like Interpretations provide a relatively local scope as to what it knows
    about the schema. As is the hierarchy grows, more and more data is is combined and aggregated together until an entire schema
    is produced. For more information on aggregation, see `AggregatedIntrospectiveIngestionComponent`.
    """

    @abstractmethod
    def gather_used_indexes(self):
        raise NotImplementedError

    @abstractmethod
    def gather_object_shapes(self):
        raise NotImplementedError

    @abstractmethod
    def gather_present_relationships(self):
        raise NotImplementedError

    def generate_graph_schema(self) -> GraphSchema:
        shapes = self.gather_object_shapes()
        relationships = self.gather_present_relationships()
        return GraphSchema(
            object_shapes=_merge_overlapping_items(shapes, shapes),
            relationships=_merge_overlapping_items(relationships, shapes),
        )


class AggregatedIntrospectiveIngestionComponent(IntrospectiveIngestionComponent):
    """A mixin to provide utilities for `IntrospectableIngestionComponent`s that are aggregations of others.

    For many types in our ingestion hierarchy, they are aggregations of subordinate components that each
    on their own provide part of the make up for a graph schema. When we look at higher order components such
    as a Interpreter, they need to merge and resolve ambiguity amongst its child ingestion components (Interpretations).

    This mixin provides that higher order aggregation logic to merge and resolve ambiguity amongst such a hierarchy.
    """

    @abstractmethod
    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        raise NotImplementedError

    def gather_used_indexes(self):
        return {
            index
            for component in self.all_subordinate_components()
            for index in component.gather_used_indexes()
        }

    def gather_object_shapes(self):
        shapes = [
            shape
            for component in self.all_subordinate_components()
            for shape in component.gather_object_shapes()
        ]
        return _merge_overlapping_items(shapes, shapes)

    def gather_present_relationships(self):
        shapes = self.gather_object_shapes()
        rels = [
            rel
            for component in self.all_subordinate_components()
            for rel in component.gather_present_relationships()
        ]
        return _merge_overlapping_items(rels, shapes)



from typing import Any, Dict, List

from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...schema.schema import AggregatedIntrospectiveIngestionComponent
from .interpretation import Interpretation


class UnhandledBranchError(ValueError):
    """Raised when a branch is not handled in a switch case."""

    def __init__(self, missing_branch_value, *args: object) -> None:
        super().__init__(
            f"'{missing_branch_value}' was not matched in switch case", *args
        )


class SwitchInterpretation(
    AggregatedIntrospectiveIngestionComponent, Interpretation, alias="switch"
):
    __slots__ = (
        "switch_on",
        "interpretations",
        "default",
        "normalization",
        "fail_on_unhandled",
    )

    @staticmethod
    def guarantee_interpretation_list_from_file_data(file_data) -> List[Interpretation]:
        if isinstance(file_data, list):
            return [
                Interpretation.from_file_data(**interpretation)
                for interpretation in file_data
            ]

        return [Interpretation.from_file_data(**file_data)]

    def __init__(
        self,
        switch_on: StaticValueOrValueProvider,
        cases: Dict[str, dict],
        default: Dict[str, Any] = None,
        normalization: Dict[str, Any] = None,
        fail_on_unhandled: bool = True,
    ):
        self.switch_on = ValueProvider.guarantee_value_provider(switch_on)
        self.interpretations = {
            field_value: self.guarantee_interpretation_list_from_file_data(
                interpretation
            )
            for field_value, interpretation in cases.items()
        }
        self.default = (
            self.guarantee_interpretation_list_from_file_data(default)
            if default
            else None
        )
        self.normalization = normalization or {}
        self.fail_on_unhandled = fail_on_unhandled

    def all_subordinate_components(self):
        for child in self.interpretations.values():
            yield from child
        if self.default:
            yield from self.default

    def interpret(self, context: ProviderContext):
        key = self.switch_on.normalize_single_value(context, self.normalization)
        interpretations = self.interpretations.get(key, self.default)

        if interpretations is None:
            if self.fail_on_unhandled:
                raise UnhandledBranchError(key)
            else:
                interpretations = []

        for interpretation in interpretations:
            interpretation.interpret(context)



from .argument_resolver import ARGUMENT_RESOLVER_REGISTRY, ArgumentResolver
from .configuration_argument_resolver import (
    ConfigurationArgumentResolver,
    get_config,
    set_config,
)
from .environment_variable_resolver import EnvironmentResolver
from .include_file_resolver import IncludeFileResolver

__all__ = (
    "ARGUMENT_RESOLVER_REGISTRY",
    "ArgumentResolver",
    "EnvironmentResolver",
    "IncludeFileResolver",
    "ConfigurationArgumentResolver",
    "get_config",
    "set_config",
)



import os.path
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Set

from ..file_io import LoadsFromYaml, SavesToYaml
from ..pipeline import Pipeline, PipelineFile, PipelineInitializationArguments
from ..schema.schema import IntrospectiveIngestionComponent


def get_default_name(file_path: Path) -> str:
    return file_path.stem


@dataclass
class PipelineDefinition(IntrospectiveIngestionComponent, SavesToYaml, LoadsFromYaml):
    """A `PipelineDefinition` represents a pipeline that can be loaded from a file.

    `PipelineDefinition` objects are used to load pipelines from files. They themselves
    are not pipelines, but rather represent the metadata needed to load a pipeline from
    a file.

    `PipelineDefinition` objects are also `IntrospectiveIngestionComponent` objects,
    meaning that they can be introspected to determine their known schema definitions.

    `PipelineDefinition` objects are also `SavesToYaml` and `LoadsFromYaml` objects,
    meaning that they can be serialized to and deserialized from YAML data.
    """

    name: str
    file_path: Path
    targets: Set[str] = frozenset()
    exclude_inherited_targets: bool = False
    annotations: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_path(cls, file_path: Path):
        """Create a `PipelineDefinition` from a file path.

        The name of the pipeline will be the stem of the file path, and the annotations
        will be empty. The pipeline will be loaded from the provided file path.

        Args:
            file_path (Path): The path to the file from which to load the pipeline.

        Returns:
            PipelineDefinition: The `PipelineDefinition` object.
        """
        return cls(get_default_name(file_path), file_path)

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Or, Schema

        return Schema(
            Or(
                str,
                {
                    Optional("name"): str,
                    "path": os.path.exists,
                    Optional("annotations"): {str: Or(str, int, float, bool)},
                    Optional("targets"): [str],
                    Optional("exclude_inherited_targets"): bool,
                },
            )
        )

    @classmethod
    def from_file_data(cls, data, parent_annotations):
        if isinstance(data, str):
            data = {"path": data}

        file_path = Path(data.pop("path"))
        name = data.pop("name", get_default_name(file_path))
        exclude_targets = data.pop("exclude_inherited_targets", False)
        annotations = data.pop("annotations", {})
        targets = data.pop("targets", [])
        return cls(
            name,
            file_path,
            set(targets),
            exclude_targets,
            {**parent_annotations, **annotations},
        )

    def to_file_data(self, verbose: bool = False):
        using_default_name = self.name == self.file_path.stem
        if using_default_name and not self.annotations and not verbose:
            return str(self.file_path)

        result = {"path": str(self.file_path)}
        if not using_default_name or verbose:
            result["name"] = self.name
        if self.annotations or verbose:
            result["annotations"] = self.annotations
        if self.targets or verbose:
            result["targets"] = self.targets

        return result

    def initialize(self, init_args: PipelineInitializationArguments) -> Pipeline:
        return PipelineFile(self.file_path).load_pipeline(init_args)

    def remove_file(self, missing_ok: bool = True):
        """Remove the file associated with this `PipelineDefinition`.

        Args:
            missing_ok (bool, optional): Whether to ignore missing files. Defaults to True.

        Raises:
            FileNotFoundError: If the file does not exist and `missing_ok` is False.
        """
        self.file_path.unlink(missing_ok=missing_ok)

    def initialize_for_introspection(self) -> Pipeline:
        return self.initialize(PipelineInitializationArguments.for_introspection())

    def gather_object_shapes(self):
        return self.initialize_for_introspection().gather_object_shapes()

    def gather_present_relationships(self):
        return self.initialize_for_introspection().gather_present_relationships()

    def gather_used_indexes(self):
        return self.initialize_for_introspection().gather_used_indexes()



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Iterable, Optional

from ..pipeline.value_providers import ProviderContext, ValueProvider


class RecordDecomposer(ABC):
    """A RecordDecomposer is responsible for looking at a record to decomposing it to sub-records to look at"""

    @abstractmethod
    def decompose_record(self, context: ProviderContext) -> Iterable[ProviderContext]:
        raise NotImplementedError

    @classmethod
    def from_iteration_arguments(cls, iteration_arguments: Optional[ValueProvider]):
        if iteration_arguments:
            return ValueProviderDecomposer(iteration_arguments)

        return WholeRecordDecomposer()


class WholeRecordDecomposer(RecordDecomposer):
    """Simply returns the original `ProviderContext`.

    `decompose_record` will take the supplied `ProviderContext` and return it as the only
    decomposed record in the set.
    """

    def decompose_record(self, context: ProviderContext) -> Iterable[ProviderContext]:
        yield context


class ValueProviderDecomposer(RecordDecomposer):
    """Iterates on the values provided from a value provider from the root context.

    `decompose_record` will take the supplied `ProviderContext` and deep copy it for each
    provided value from the value provider when called on the supplied context.
    """

    def __init__(self, value_provider: ValueProvider) -> None:
        self.value_provider = value_provider

    def decompose_record(self, context: ProviderContext) -> Iterable[ProviderContext]:
        for sub_document in self.value_provider.many_values(context):
            sub_context = deepcopy(context)
            sub_context.document = sub_document
            yield sub_context



import time
from decimal import Decimal
from logging import getLogger
from typing import Any, AsyncGenerator

from ...extractor import Extractor
from .credential_utils import AwsClientFactory

ATHENA_STATE_RUNNING = "RUNNING"
ATHENA_STATE_FAILED = "FAILED"
ATHENA_STATE_CANCELLED = "CANCELLED"
ATHENA_STATE_QUEUED = "QUEUED"

PENDING_ATHENA_STATES = {ATHENA_STATE_QUEUED, ATHENA_STATE_RUNNING}
BAD_ATEHNA_STATES = {ATHENA_STATE_FAILED, ATHENA_STATE_CANCELLED}


CONVERTERS = {
    "tinyint": int,
    "smallint": int,
    "integer": int,
    "bigint": int,
    "double": float,
    "float": float,
    "decimal": Decimal,
    "char": str,
    "string": str,
    "boolean": lambda v: v == "true",
}


def leave_untuched(value):
    return value


class AthenaRowConverter:
    def __init__(self, column_meta) -> None:
        self.column_meta = column_meta

    def convert_row(self, row):
        return {
            column_meta["Name"]: self.convert_value(column_meta, value)
            for column_meta, value in zip(self.column_meta, row["Data"])
        }

    def convert_value(self, column_metadata, column_value):
        raw_value = column_value.get("VarCharValue")
        if raw_value is None:
            return None
        convert = CONVERTERS.get(column_metadata["Type"], leave_untuched)
        return convert(raw_value)


class AthenaExtractor(Extractor):
    @classmethod
    def from_file_data(
        cls,
        query: str,
        database: str,
        workgroup: str,
        output_location: str,
        poll_interval_seconds: int = 1,
        page_size: int = 500,
        **aws_client_args,
    ):
        client = AwsClientFactory(**aws_client_args).make_client("athena")
        return cls(
            query=query,
            database=database,
            workgroup=workgroup,
            output_location=output_location,
            poll_interval_seconds=poll_interval_seconds,
            page_size=page_size,
            client=client,
        )

    def __init__(
        self,
        query: str,
        database: str,
        workgroup: str,
        output_location: str,
        client,
        poll_interval_seconds: int,
        page_size: int,
    ) -> None:
        self.query = query
        self.database = database
        self.workgroup = workgroup
        self.output_location = output_location
        self.client = client
        self.poll_interval_seconds = poll_interval_seconds
        self.page_size = page_size
        self.logger = getLogger(self.__class__.__name__)

    def execute_query(self):
        result = self.client.start_query_execution(
            QueryString=self.query,
            QueryExecutionContext={"Database": self.database},
            ResultConfiguration={"OutputLocation": self.output_location},
            WorkGroup=self.workgroup,
        )
        self.query_execution_id = result["QueryExecutionId"]
        self.logger.debug(
            "Athena Query Started", dict(query_execution_id=self.query_execution_id)
        )

    def get_query_status(self):
        result = self.client.get_query_execution(
            QueryExecutionId=self.query_execution_id
        )
        return result["QueryExecution"]["Status"]["State"]

    def await_query_completion(self):
        while (status := self.get_query_status()) in PENDING_ATHENA_STATES:
            time.sleep(self.poll_interval_seconds)

        if status in BAD_ATEHNA_STATES:
            raise RuntimeError(f"Failed with bad athena query state: {status}")

    def get_result_paginator(self):
        paginator = self.client.get_paginator("get_query_results")
        return paginator.paginate(
            QueryExecutionId=self.query_execution_id,
            PaginationConfig={"PageSize": self.page_size},
        )

    def page_results_and_get_rows_with_metadata(self):
        for page in self.get_result_paginator():
            column_meta = page["ResultSet"]["ResultSetMetadata"]["ColumnInfo"]
            for row in page["ResultSet"]["Rows"]:
                yield row, column_meta

    def convert_data_types_of_rows_based_on_headers(self, rows_with_meta):
        row, page_meta = next(rows_with_meta)
        converter = AthenaRowConverter(page_meta)
        for row, _ in rows_with_meta:
            yield converter.convert_row(row)

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        self.execute_query()
        self.await_query_completion()
        rows_with_meta = self.page_results_and_get_rows_with_metadata()
        for result in self.convert_data_types_of_rows_based_on_headers(rows_with_meta):
            yield result



from .extractor import StreamConnector, StreamExtractor, StreamRecordFormat
from .kafka import KafkaStreamConnector

__all__ = (
    "StreamConnector",
    "StreamRecordFormat",
    "StreamExtractor",
    "KafkaStreamConnector",
)



from collections import defaultdict

from ..schema import GraphSchema
from .schema_printer import SchemaPrinter


class LargeLanguageModelSchemaPrinter(SchemaPrinter, alias="genaillm"):
    def return_nodes_props(self, schema):
        return {
            str(node_shape.object_type): node_shape.property_names()
            for node_shape in schema.known_node_types()
        }

    def return_rels_props(self, schema):
        return {
            str(rel_shape.object_type): rel_shape.property_names()
            for rel_shape in schema.known_relationship_types()
        }

    def return_rels(self, schema):
        rels = defaultdict(list)
        for elem in schema.relationships:
            from_node = elem.from_object_type
            to_node = elem.to_object_type
            rel_name = elem.relationship_type
            rels[str(from_node)] = defaultdict(list)
            rels[str(from_node)][str(rel_name)].append(str(to_node))
        return rels

    def print_schema_to_string(self, schema: GraphSchema) -> str:
        representation = str(self.return_nodes_props(schema))
        representation += ". "
        representation += str(self.return_rels_props(schema))
        representation += ". "
        representation += str(self.return_rels(schema))
        return representation



from ...interpreting import Interpreter
from ...schema.indexes import KeyIndex
from ..project import Project
from .audit import Audit


class AuditReferentialIntegrity(Audit):
    name = "refs"
    description = "Audit the correctness of references between nodes of the project"

    async def run(self, project: Project):
        all_interpreters = project.dig_for_step_of_type(Interpreter)
        node_types_so_far = {}

        for definition, step_index, interpreter in all_interpreters:
            for index in interpreter.gather_used_indexes():
                if not isinstance(index, KeyIndex):
                    continue

                if index.type in node_types_so_far:
                    existing = node_types_so_far[index.type]
                    existing_index, existing_definition, existing_step_index = existing
                    if index != existing_index:
                        current_index_fields = ", ".join(index.identity_keys)
                        existing_index_fields = ", ".join(existing_index.identity_keys)
                        self.failure(
                            f"Regarding the type '{index.type}':\n"
                            f"\tthe Interpreter in '{definition.file_path}' step '{step_index}' wants to use an index on the fields '{current_index_fields}' WHILE\n"
                            f"\tthe Interpreter in '{existing_definition.file_path}' step '{existing_step_index}' wants to use an index on the fields '{existing_index_fields}'"
                        )
                else:
                    node_types_so_far[index.type] = (index, definition, step_index)

        if self.failure_count == 0:
            self.success("All Object Types Have Consistent Referential Integrity")



from typing import Any, Iterable, Type

from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import ValueProvider


class VariableValueProvider(ValueProvider):
    """A `ValueProvider` that uses a variable to extract values from a document."""

    __slots__ = ("variable_name",)

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!variable",
            lambda loader, node: cls(loader.construct_scalar(node)),
        )

    def __init__(self, variable_name: str) -> None:
        self.variable_name = variable_name

    def single_value(self, context: ProviderContext) -> Any:
        return context.variables.get(self.variable_name)

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        value = self.single_value(context)
        if value is None:
            return []
        return value if isinstance(value, list) else [value]


SafeDumper.add_representer(
    VariableValueProvider,
    lambda dumper, mapping: dumper.represent_scalar("!variable", mapping.variable_name),
)



from ...pluggable import Pluggable
from ..project import Project
from .audit_printer import AuditPrinter


class Audit(Pluggable):
    name = "unnamed"
    description = "Checks something about a project"
    entrypoint_name = "audits"

    def __init__(self, printer: AuditPrinter) -> None:
        self.printer = printer
        self.success_count = 0
        self.warning_count = 0
        self.failure_count = 0

    async def run(self, project: Project) -> None:
        pass

    def success(self, message: str) -> None:
        """Print a success message.

        The total number of successes will be tracked and is accessible via the `success_count` attribute.

        Args:
            message: The message to print.
        """
        self.success_count += 1
        self.printer.print_success(message)

    def failure(self, message: str) -> None:
        """Print a failure message.

        The total number of failures will be tracked and is accessible via the `failure_count` attribute.
        Non zero failure counts will cause the audit to fail and the command line to exit with a non-zero exit
        code. Specifically, the exit code will be the number of failures.

        Args:
            message: The message to print.
        """
        self.failure_count += 1
        self.printer.print_failure(message)

    def warning(self, message: str) -> None:
        """Print a warning message.

        The total number of warnings will be tracked and is accessible via the `warnings_count`
        attribute.
        """
        self.warning_count += 1
        self.printer.print_warning(message)



from abc import ABC, abstractmethod

from ..pluggable import Pluggable
from ..subclass_registry import SubclassRegistry
from .copy import TypeRetriever
from .query_executor import QueryExecutor
from .query_executor_with_statistics import QueryExecutorWithStatistics

DATABASE_CONNECTOR_SUBCLASS_REGISTRY = SubclassRegistry()


@DATABASE_CONNECTOR_SUBCLASS_REGISTRY.connect_baseclass
class DatabaseConnector(ABC, Pluggable):
    entrypoint_name = "databases"

    @classmethod
    def from_database_args(
        cls, database: str = "neo4j", **database_args
    ) -> "DatabaseConnector":
        DatabaseConnector.import_all()

        return DATABASE_CONNECTOR_SUBCLASS_REGISTRY.get(database).from_file_data(
            **database_args
        )

    @classmethod
    def from_file_data(cls, **kwargs):
        return cls(**kwargs)

    @abstractmethod
    def make_query_executor(self) -> QueryExecutor:
        raise NotImplementedError

    @abstractmethod
    def make_type_retriever(self) -> TypeRetriever:
        raise NotImplementedError

    def get_query_executor(self, collect_stats: bool = True):
        connector = self.make_query_executor()
        if collect_stats:
            connector = QueryExecutorWithStatistics(connector)
        return connector

    def get_type_retriever(self):
        return self.make_type_retriever()
